/*
*	lr: list directories recursively
*	version: 2.3b
*	date: 2021/11/05
*	auther: Kenar(Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*
*	compiling under plan9port
*		mk -f mkfile_p9p
*/

#include <u.h>
#include <libc.h>
#include <bio.h>

#ifdef PLAN9PORT
#include <libgen.h>
#include <sys/stat.h>
#include <dirent.h>
/* I added some DMXXX. look libc.h in p9p */
#define DMBLOCK 0x00800000	/* same as DMDEVICE */
#define DMCHAR  0x00400000	/* we have room for new definition */
#define SPECIAL (DMMOUNT|DMDEVICE|DMNAMEDPIPE|DMSOCKET)  /* special files in unix */
/* stime may conflict on some UNIXs */
#endif

#define DBG if(0)

void
usage(void)
{
	fprint(2,"usage: lr [-flQpqinaL] [-d depth] [-D depth] [-t time]  path ...\n"
	"lr lists directory recursively\n"
	"options:\n"
	"	-f: list only files\n"
	"	-l: long format\n"
	"	-Q: unquoted format, instead TAB as separator (used only in combination with '-l')\n"
	"	-at time: show files accessed since the time (seconds since the unix epoch)\n"
	"	-d depth: search directory upto depth (depth >= 0)\n"
	"	-D depth: search directory of depth (depth >= 0)\n"
	"	-i: show date using iso-8601 style\n"
	"	-n: show date using unix epoch\n"
	"	-q: show qid\n"
	"	-t time: show files modified since the time (seconds since the unix epoch)\n"
	"	-L: follow links (for unix)\n"
	"	-p: show only path elemnts followed given arguments\n"
	"version: 2.2\n"
	);
	exits("usage");
}

enum
{
	LFMT = 1,
	nFMT = 2,
	QFMT = 4,
	iFMT = 8,
	qFMT = 16,
};

char *mods[]={"---","--x","-w-","-wx","r--","r-x","rw-","rwx"};

int baselen = 0;
int mode = 0;
Biobuf	out;
ulong stime = 0;
int pflag = 0;
int aflag = 0; /* accessed time */
int fflag = 0;
int Dflag = 0;
int Lflag = 0;

char*
modefmt(ulong m)
{
	static char buf[32];
	char *e, *s;
	s = buf;
	e = s + sizeof(buf);
#ifdef PLAN9PORT
	s[0] = '-';
	if(m & DMDIR)
		s[0] = 'd';
	if(m & DMNAMEDPIPE)
		s[0] = 'p';
	if(m & DMSOCKET)
		s[0] = 's';
	/* I have added DMBLOCK and DMCHAR */
	if(m & DMBLOCK)
		s[0] = 'b';
	if(m & DMCHAR)
		s[0] = 'c';
	if(!Lflag && (m & DMSYMLINK))
		s[0] = 'l';
	s += 1;
#else
	s[0] = s[1] = '-';
	if(m & DMDIR)
	s[0] = 'd';
	if(m & DMAPPEND)
		s[0] = 'a';
	if(m & DMEXCL)
		s[1] = 'l';
	s += 2;
#endif
	s = seprint(s,e, "%s", mods[(m & 0700) >> 6]);
	s = seprint(s,e, "%s", mods[(m & 070) >> 3]);
	seprint(s,e, "%s", mods[m & 07]);

#ifdef PLAN9PORT
	/* look $P9P/include/libc.h
	 * we have DMSETUID and DMSETGID
	 * however this is no effect. probably russ forgot implementation 
	 * look $P9P/src/lib9/_p9dir.c */
	if(m & DMSETUID)
		buf[3] = 's';
	if(m & DMSETGID)
		buf[6] = 's';
#endif
	return buf;
}

char *
datefmt(ulong t)
{
	static char buf[48];
	Tm *tm;
	char sep = '/';

	if(mode & nFMT){
		snprint(buf,sizeof buf, "%lud", t);
		return buf;
	}
	tm = localtime(t);
	if(mode & iFMT)
		sep = '-';
	snprint(buf,sizeof buf,"%04d%c%02d%c%02d %02d:%02d:%02d",
		tm->year + 1900,sep,tm->mon + 1,sep,tm->mday, tm->hour, tm->min, tm->sec);
	return buf;
}

char *
qidfmt(Qid qid)
{
	static char buf[48];
	snprint(buf,sizeof buf, "%.16llux %.5lud %.2ux", qid.path, qid.vers, qid.type);
	return buf;
}


/*
*	don't use d->name, the name is in path
*/
void
ls(char *path, Dir *d)
{
	char *fmt, *trail="";
	int n = 0;
#ifdef PLAN9PORT
	struct stat sb;
	if((d->mode & DMSYMLINK) && (mode & LFMT) && Lflag){
		if(stat(path, &sb)<0)
			return;
		if((sb.st_mode & S_IFMT) == S_IFREG){ /* regular file */
			d->mode &= ~DMSYMLINK;
			d->length = (vlong)sb.st_size;
		}
		if((sb.st_mode & S_IFMT) == S_IFDIR){ /* directory */
			d->mode &= ~DMSYMLINK;
			d->mode |= DMDIR;
			d->length = 0;
		}
	}
	
	if(d->mode & DMDEVICE){
		if(stat(path, &sb)<0)
			return;
		if((sb.st_mode & S_IFMT) == S_IFCHR){
			d->mode &= ~DMDEVICE;
			d->mode |= DMCHAR;
		}
	}
#endif
	path += baselen;
	if(mode & qFMT){
		fmt = "%s ";
		Bprint(&out,fmt, qidfmt(d->qid));
		n++;
	}

	if(mode & LFMT){
		fmt = "%s %s %s %10llud %s ";
		Bprint(&out,fmt, modefmt(d->mode),
		d->uid, d->gid, d->length, datefmt(d->mtime));
		n++;
	}

	if((n > 0) && (mode & QFMT)){
		Bungetc(&out);
		Bputc(&out,'\t');
	}
	if((d->mode & DMDIR) && !(mode & LFMT))
		trail = "/";
	fmt = (mode & QFMT)?"%s%s\n":"%q%s\n";
	Bprint(&out,fmt, path,trail);
}

int
compar(Dir *a, Dir *b)
{
	return strcmp(a->name, b->name);
}


#ifdef PLAN9PORT

typedef struct List List;
struct List{
	uvlong	*qpath;
	List	*next;
};

List *pathlist;

/* putmark appends path to pl if absent
 * if already exist it returns -1 */
int
putmark(List **pl, uvlong qidpath)
{
	List *p, *q;
	q = nil;
	p = *pl;
	while(p){
		if(p->qpath == qidpath)
			return -1;
		q = p;
		p = p->next;
	}
	/* then p is nil */
	p = malloc(sizeof(List));
	if(p == nil)
		sysfatal("out of memory");
	if(q)
		q->next = p;
	else
		*pl = p;
	p->qpath = qidpath;
	p->next = nil;
	return 0;
}
#endif

void
list(char *path, char *ep, int depth)
{
	Dir *db;
	int fd, i,n;
	ulong m;
#ifdef PLAN9PORT
	char *p;
	Dir *db0;
	struct stat sb;
	mode_t ms;
#endif

	/* NB: this dirstat() is from russ
	 * therefore not be reflected my change in _mp9dir.c */
	db = dirstat(path); /* path may be for file or for dir */
	if(db == nil){
		Bflush(&out);
		fprint(2, "# dirstat %s: %r\n", path);	
		return;
	}
	m = db->mode;

#ifdef PLAN9PORT
	/* use _mp9dir.c instead of _p9dir.c */
	/* NB: we have two stat():
	 *   unix stat(path,&sb)
	 *   p9p stat(path,edir,nedir)
	 * these two may conflict?
	 * Russ says:
	 *   extern	int	stat(char*, uchar*, int); give up
	 * in $PLAN9/include/libc.h
	 * if info of symlink itselt is required,
	 *   use lstat(path, &sb) instead of stat(path, &sb)
	 * the lstat() is only for unix
	 */
	ms = 0;
	if(m & DMSYMLINK){
		if(stat(path, &sb) < 0){
			/* 
			 * it seems stat() detects cyclic link,
			 * not only broken link
			 * the error happens:
			 *   path is non existent
			 *   linked path is non existent
			 */
			Bflush(&out);
			fprint(2,"# %s: broken or cyclic link\n", path);
			free(db);
			return;
		}
		/* 
		  *  the st_mode is linked one, that is
		  *     link foo -> bar then
		  *  the path is for foo but the contents is for bar
		  *  thefore st_mode should be that of non-symlink file
		  */
		ms = sb.st_mode;
		if(putmark(&pathlist,db->qid.path) < 0){
			Bflush(&out);
			fprint(2,"# %s: cyclic directory link\n", path);
			free(db);
			return;
		}
		/* here ms != 0 if symlink */
	}

	if(1
	&& (db->mtime >= stime || (aflag && db->atime >= stime))
	&& (!Dflag || Dflag && (depth == 0))
	&& (!fflag || (!(m & DMDIR) && (!ms || ((ms&S_IFMT) != S_IFDIR)))))
			ls(path, db);
	free(db);
	if(!(m & DMDIR) && (!ms || ((ms&S_IFMT) != S_IFDIR)))
			return;
	/* here the path  should be to dir or link to dir. that is,
	*  	(m & DMDIR) || (ms && ((ms&S_IFMT) == S_IFDIR)))
	*/
#else
	if(1
	&& (db->mtime >= stime || (aflag && db->atime >= stime))
	&& (!Dflag || Dflag && (depth == 0))
	&& (!fflag || !(m & DMDIR)))
			ls(path, db);
	free(db);
	if(!(m & DMDIR))
		return;	
	/* here the path  should be to dir */
#endif

	if(depth == 0)
		return;
	depth--;

	fd = open(path, OREAD);
	if(fd < 0){
		Bflush(&out);
		fprint(2, "# %s: unreadable\n", path);
		return;
	}
	n = dirreadall(fd, &db);
	close(fd);
	if(n < 0){
		/* I believe db should not be allocated in case of error,
		 * so free(db) is not here */
		Bflush(&out);
		fprint(2, "# dirreadall %s: %r\n", path);
		return;
	}
	
#ifdef PLAN9PORT
	/* remove special files in the db */
	if(fflag){
		int i,k;
		for(i = 0, k = 0; i < n; i++){
			if(!(db[i].mode & SPECIAL)){
				db[k] = db[i];
				k++;
			}
		}
		n = k;
	}
#endif

	/* sort by name */
	qsort(db, n, sizeof(Dir), (int (*)(const void*, const void*))compar);// for unix
	for(i = 0; i < n; i++){
		if(db[i].name[0] == 0){
			/*  this case should not happen.
			*	but old cwfs (before 2013) allowed names that break UTF-8
			*	therefore we need some checks	-Kenar- */
			continue;
		}
		if(db[i].atime < stime)
			continue;
		m = strlen(path);
		if(path[m-1] == '/')
			seprint(path + m - 1,ep,"/%s", db[i].name);
		else
			seprint(path + m,ep,"/%s", db[i].name);
		list(path, ep, depth);
		path[m] = 0;
	}
	free(db);
	return;
}


void main(int argc, char *argv[])
{
	char path[4096];
	char *ep, *ap,*p;
	int depth = -1;
	ARGBEGIN{
	case 'a':	aflag = 1; break;
	case 'D':	Dflag++; /* continue next case */
	case 'd':
		ap = EARGF(usage());
		depth = atoi(ap);
		break;
	case 'f':	fflag++; break;
	case 'l':	mode |= LFMT; break;
	case 'n':	mode |= nFMT; break;
	case 'Q':	mode |= QFMT; break;
	case 'i':	mode |= iFMT; break;
	case 'q':	mode |= qFMT; break;
	case 'p':	pflag = 1; break;
	case 't':
		ap = EARGF(usage());
		stime = (ulong)atol(ap);
		break;
	case 'L':	Lflag = 1; break;
	default: usage();
	}ARGEND

	quotefmtinstall();
	Binit(&out, 1, OWRITE);
	ep = path + sizeof path;
	if(*argv == nil)
		usage();
	while(*argv){
		strecpy(path, ep, *argv);
		p = path;
		if(pflag)
			baselen = strlen(path);
		if(strcmp(path,"/") == 0)
			sysfatal("path \"/\" is not allowed");
		if(path[strlen(p)-1] == '/')
			path[strlen(p)-1] = 0;
		if(p[0] == '.' && p[1] == '/')
			p += 2;
		list(p, ep, depth);
		argv++;
	}
	exits(nil);
}
